import { MetaDataType, NullableProperties, PageNumberPagination, UUID } from '.'

interface SearchFacets {
  element_type: string
  element_worker: string
  transcription_worker: string
  classification_name: string
  classification_worker: string
  metadata_name: string
  metadata_type: MetaDataType
  metadata_worker: string
  entity_type: string
  entity_worker: string
}

export type SearchSort = 'relevance' | 'element_name'

export type SearchSource = 'element' | 'entity' | 'metadata' | 'transcription'

// A search query allows filtering on all facets, since that's what they are for, so we do not have to repeat them
export interface SearchQuery extends Partial<SearchFacets> {
  query: string
  only_facets?: boolean
  sort?: SearchSort
  /**
   * List of sources of search results.
   *
   * Since this is a query parameter, this can be a single source and not an array, and will be converted to an array by the backend.
   */
  sources?: SearchSource | SearchSource[]
}

export type SearchResult = {
  id: UUID
  element_id: UUID
  element_text: string
  element_type: string
  element_worker: string | null
  element_image: string | null
}
// An element's parent, when there is one, always has all its properties
& NullableProperties<{
  parent_id: UUID
  parent_name: string
  parent_type: string
}>
& NullableProperties<{
  transcription_id: UUID
  transcription_text: string
  /*
   * The confidence and worker are still nullable even when there is a transcription,
   * but they can only be set if the transcription ID and text are set.
   */
  transcription_confidence: number | null
  transcription_worker: string | null
}>
& NullableProperties<{
  classification_id: UUID
  classification_name: string
  classification_confidence: number | null
  classification_worker: string | null
}>
& NullableProperties<{
  metadata_id: UUID
  metadata_name: string
  metadata_text: string
  metadata_type: MetaDataType
  metadata_worker: string | null
}>
& NullableProperties<{
  entity_id: UUID
  entity_text: string
  entity_type: string
  entity_worker: string | null
}>

// When only_facets is set to false, the results will be omitted.
type SearchPagination<T> = Omit<PageNumberPagination<T>, 'results'> & { results: T[] | null }

export type SearchFacetsResponse = {
  // Each search facet name is always defined
  [Facet in keyof SearchFacets]: {
    // The keys only are of the facet's type, and no key is guaranteed to be present
    [value in SearchFacets[Facet]]?: number
  }
}

export interface SearchResponse extends SearchPagination<SearchResult> {
  facets: SearchFacetsResponse
}
