import axios from 'axios'
import { assert } from 'chai'
import { createPinia, setActivePinia } from 'pinia'
import { mutations, getters } from '@/store/elements.js'
import { useJobsStore } from '@/stores'
import { assertRejects, FakeAxios } from '../testhelpers.js'
import store from './index.spec.js'
import { elementNeighborsSample, jobsSample } from '../samples.js'

describe('elements', () => {
  describe('mutations', () => {
    describe('set', () => {
      it('adds unknown elements', () => {
        const state = {
          elements: {
            element1: { id: 'element1', name: 'Element 1' }
          }
        }
        mutations.set(state, { id: 'element2', name: 'Element 2' })
        assert.deepStrictEqual(state, {
          elements: {
            element1: { id: 'element1', name: 'Element 1' },
            element2: { id: 'element2', name: 'Element 2' }
          }
        })
      })

      it('merges with existing elements', () => {
        const state = {
          elements: {
            element1: {
              id: 'element1',
              name: 'Element 1'
            }
          }
        }
        mutations.set(state, {
          id: 'element1',
          has_children: true
        })
        assert.deepStrictEqual(state, {
          elements: {
            element1: {
              id: 'element1',
              name: 'Element 1',
              has_children: true
            }
          }
        })
      })
    })

    describe('bulkSet', () => {
      it('stores elements in bulk', () => {
        const state = {
          elements: {
            element1: {
              id: 'element1',
              name: 'Element 1'
            }
          },
          links: {
            element1: {
              parents: [],
              children: []
            }
          }
        }
        mutations.bulkSet(state, [
          { id: 'element1', has_children: true },
          { id: 'element2', name: 'Element 2' }
        ])
        assert.deepStrictEqual(state, {
          elements: {
            element1: {
              id: 'element1',
              name: 'Element 1',
              has_children: true
            },
            element2: {
              id: 'element2',
              name: 'Element 2'
            }
          },
          links: {
            element1: {
              parents: [],
              children: []
            },
            element2: {
              parents: [],
              children: []
            }
          }
        })
      })
    })

    describe('addChildren', () => {
      it('initializes empty links', () => {
        const state = {
          elements: {
            element1: { id: 'element1', name: 'Element 1' }
          },
          links: {}
        }

        mutations.addChildren(state, {
          element: 'element1',
          children: [
            {
              id: 'element2',
              name: 'Element 2'
            }
          ]
        })

        assert.deepStrictEqual(state, {
          elements: {
            element1: { id: 'element1', name: 'Element 1' },
            element2: { id: 'element2', name: 'Element 2' }
          },
          links: {
            element1: {
              parents: [],
              children: ['element2']
            },
            element2: {
              parents: ['element1'],
              children: []
            }
          }
        })
      })

      it('adds to existing links', () => {
        const state = {
          elements: {
            element1: { id: 'element1', name: 'Element 1' },
            element2: { id: 'element2', name: 'Element 2' }
          },
          links: {
            element1: {
              parents: [],
              children: ['element2']
            },
            element2: {
              parents: ['element1'],
              children: []
            }
          }
        }

        mutations.addChildren(state, {
          element: 'element2',
          children: [
            {
              id: 'element3',
              name: 'Element 3'
            }
          ]
        })

        assert.deepStrictEqual(state, {
          elements: {
            element1: { id: 'element1', name: 'Element 1' },
            element2: { id: 'element2', name: 'Element 2' },
            element3: { id: 'element3', name: 'Element 3' }
          },
          links: {
            element1: {
              parents: [],
              children: ['element2']
            },
            element2: {
              parents: ['element1'],
              children: ['element3']
            },
            element3: {
              parents: ['element2'],
              children: []
            }
          }
        })
      })

      it('ignores falsy values and empty arrays', () => {
        const state = {
          elements: {
            element1: { id: 'element1', name: 'Element 1' }
          },
          links: {
            element1: {
              parents: [],
              children: []
            }
          }
        }

        mutations.addChildren(state, { element: 'element1', children: [] })
        mutations.addChildren(state, { element: 'element1', children: false })
        mutations.addChildren(state, { element: 'element1', children: null })
        mutations.addChildren(state, { element: 'element1' })

        // Nothing should change
        assert.deepStrictEqual(state, {
          elements: {
            element1: { id: 'element1', name: 'Element 1' }
          },
          links: {
            element1: {
              parents: [],
              children: []
            }
          }
        })
      })
    })

    describe('addNeighbors', () => {
      it('add neighbors to an element', () => {
        const state = { neighbors: {} }
        const neighbors = [
          {
            previous: { id: 'elementid' },
            next: { id: 'elementid' },
            path: [
              { id: 'folderid' }
            ],
            ordering: 42
          }
        ]
        mutations.addNeighbors(state, { element: 'element1', neighbors })
        assert.deepStrictEqual(state, { neighbors: { element1: neighbors } })
      })

      it('removes unknown parents', () => {
        const payload = [
          {
            previous: { id: 'elementid' },
            next: { id: 'elementid' },
            path: [
              { id: 'folderid' },
              null
            ],
            ordering: 42
          }
        ]
        const expected = [
          {
            previous: { id: 'elementid' },
            next: { id: 'elementid' },
            path: [
              { id: 'folderid' }
            ],
            ordering: 42
          }
        ]
        const state = { neighbors: {} }

        assert.throws(
          () => mutations.addNeighbors(state, { element: 'element1', neighbors: payload }),
          'Unknown parent elements were found in neighbors'
        )
        assert.deepStrictEqual(state, { neighbors: { element1: expected } })
      })

      it('handles empty values', () => {
        const state = { neighbors: { element1: { something: 'blah' } } }
        mutations.addNeighbors(state, { element: 'element1', neighbors: {} })
        assert.deepStrictEqual(state, { neighbors: { element1: {} } })
      })

      it('handles null values', () => {
        const state = { neighbors: { element1: { something: 'blah' } } }
        mutations.addNeighbors(state, { element: 'element1', neighbors: null })
        assert.deepStrictEqual(state, { neighbors: { element1: null } })
      })
    })

    describe('updatedElements', () => {
      it('toggles the elementsUpdate boolean', () => {
        const state = { elementsUpdate: true }
        mutations.updatedElements(state)
        assert.strictEqual(state.elementsUpdate, false)
        mutations.updatedElements(state)
        assert.strictEqual(state.elementsUpdate, true)
      })
    })

    it('reset', () => {
      const state = {
        elements: {
          element1: {
            id: 'element1',
            name: 'Element 1'
          }
        },
        parents: {
          element1: ['element2']
        },
        neighbors: {
          element1: elementNeighborsSample
        },
        links: {
          element1: {
            children: ['element2']
          }
        },
        childrenPagination: {
          element1: {
            count: 1,
            number: 1
          }
        },
        elementsUpdate: true
      }
      mutations.reset(state)
      assert.deepStrictEqual(state, {
        elements: {},
        parents: {},
        neighbors: {},
        links: {},
        childrenPagination: {},
        elementsUpdate: false
      })
    })

    describe('addClassification', () => {
      it('Add a new classification on an element', () => {
        const state = {
          elements: {
            element1: { id: 'element1' }
          }
        }
        mutations.addClassification(state, {
          elementId: 'element1',
          value: { id: 'classif1', state: 'pending' }
        })
        assert.deepStrictEqual(state, {
          elements: {
            element1: {
              id: 'element1',
              classifications: [
                { id: 'classif1', state: 'pending' }
              ]
            }
          }
        })
      })
    })

    describe('removeClassification', () => {
      it('Removes a classification', () => {
        const state = {
          elements: {
            element1: {
              id: 'element1',
              classifications: [
                { id: 'classif1', state: 'pending' },
                { id: 'classif2', state: 'pending' }
              ]
            }
          }
        }
        mutations.removeClassification(state, {
          elementId: 'element1',
          value: { id: 'classif2' }
        })
        assert.deepStrictEqual(state, {
          elements: {
            element1: {
              id: 'element1',
              classifications: [
                { id: 'classif1', state: 'pending' }
              ]
            }
          }
        })
      })
    })

    describe('updateClassification', () => {
      it('Updates the classification of an element', () => {
        const state = {
          elements: {
            element1: {
              id: 'element1',
              classifications: [
                { id: 'classif1', state: 'pending' },
                { id: 'classif2', state: 'pending' }
              ]
            }
          }
        }
        mutations.updateClassification(state, {
          elementId: 'element1',
          value: { id: 'classif1', state: 'validated' }
        })
        assert.deepStrictEqual(state, {
          elements: {
            element1: {
              id: 'element1',
              classifications: [
                { id: 'classif1', state: 'validated' },
                { id: 'classif2', state: 'pending' }
              ]
            }
          }
        })
      })
    })
  })

  describe('actions', () => {
    let mock, jobsStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      jobsStore = useJobsStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.reset()
      jobsStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('get', () => {
      it('retrieves an element', async () => {
        const element1 = {
          id: 'element1',
          name: 'Element 1'
        }
        mock.onGet('/element/element1/').reply(200, element1)

        const element = await store.dispatch('elements/get', { id: 'element1' })

        assert.deepStrictEqual(store.history, [
          { action: 'elements/get', payload: { id: 'element1' } },
          { mutation: 'elements/set', payload: element1 }
        ])
        assert.deepStrictEqual(store.state.elements.elements, { element1 })
        assert.deepStrictEqual(element, element1)
      })

      it('handles errors', async () => {
        mock.onGet('/element/element1/').reply(500)
        await assertRejects(async () => { await store.dispatch('elements/get', { id: 'element1' }) })
        assert.deepStrictEqual(store.history, [
          { action: 'elements/get', payload: { id: 'element1' } },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
      })
    })

    describe('create', () => {
      it('creates an element with a parent', async () => {
        mock.onPost('/elements/create/').reply(201, { id: 'created_id', name: 'new_name', zone: null })
        store.state.elements = {
          elements: {
            element_id: { id: 'element_id', name: 'element' },
            parent_id: { id: 'parent_id', name: 'parent' }
          },
          elementsUpdate: true,
          links: {
            element_id: { parents: ['parent_id'], children: [] },
            parent_id: { parents: [], children: ['element_id'] }
          }
        }

        await store.dispatch('elements/create', { name: 'new_name', parent: 'parent_id' })

        assert.deepStrictEqual(store.history, [
          { action: 'elements/create', payload: { name: 'new_name', parent: 'parent_id' } },
          {
            mutation: 'elements/set',
            payload: {
              id: 'created_id',
              name: 'new_name',
              zone: null
            }
          },
          {
            mutation: 'elements/addChildren',
            payload: {
              element: 'parent_id',
              children: [
                {
                  id: 'created_id',
                  name: 'new_name',
                  zone: null
                }
              ]
            }
          },
          {
            mutation: 'elements/updatedElements'
          }
        ])

        assert.deepStrictEqual(store.state.elements, {
          elements: {
            created_id: { id: 'created_id', name: 'new_name', zone: null },
            element_id: { id: 'element_id', name: 'element' },
            parent_id: { id: 'parent_id', name: 'parent' }
          },
          elementsUpdate: false,
          links: {
            created_id: { parents: ['parent_id'], children: [] },
            element_id: { parents: ['parent_id'], children: [] },
            parent_id: { parents: [], children: ['element_id', 'created_id'] }
          }
        })
      })

      it('handles errors', async () => {
        mock.onPost('/elements/create/').reply(400, { name: ['Sorry, I ate it.'] })

        await assertRejects(async () => {
          await store.dispatch('elements/create', { name: 'pizza' })
        })

        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/create',
            payload: { name: 'pizza' }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 400'
            }
          }
        ])
      })
    })

    describe('patch', () => {
      it('updates an element', async () => {
        mock.onPatch('/element/element1/').reply(200, { id: 'element1', name: 'new_name' })
        store.state.elements = {
          elements: {
            element1: { id: 'element1', name: 'old_name' }
          },
          links: {},
          childrenPagination: {}
        }
        await store.dispatch('elements/patch', { id: 'element1', name: 'new_name' })
        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/patch', payload: { id: 'element1', name: 'new_name' }
          },
          {
            mutation: 'elements/set',
            payload: {
              id: 'element1',
              name: 'new_name'
            }
          },
          {
            mutation: 'elements/updatedElements'
          }
        ])
        assert.deepStrictEqual(store.state.elements.elements, {
          element1: { id: 'element1', name: 'new_name' }
        })
      })
    })

    describe('delete', () => {
      it('deletes an element', async () => {
        mock.onDelete('/element/element/').reply(200)
        mock.onGet('/jobs/').reply(200, jobsSample)

        await store.dispatch('elements/delete', { id: 'element' })
        await store.actionsCompleted()

        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/delete',
            payload: { id: 'element' }
          },
          {
            mutation: 'elements/remove',
            payload: 'element'
          },
          {
            mutation: 'notifications/notify',
            payload: { type: 'success', text: 'Element deletion has been scheduled.' }
          }
        ])
        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })
    })

    describe('nextChildren', () => {
      it('loads multiple pages', async () => {
        const element1 = { id: 'element1', name: 'Element 1' }
        const element2 = { id: 'element2', name: 'Element 2' }
        const element3 = { id: 'element3', name: 'Element 3' }
        store.state.elements = {
          elements: {
            element1
          },
          links: {},
          childrenPagination: {}
        }
        const page1pagination = {
          count: 3,
          previous: null,
          next: '/elements/element1/children/?page=2',
          number: 1
        }
        mock.onGet('/elements/element1/children/', { params: { page: 1 } }).reply(200, {
          ...page1pagination,
          results: [element2]
        })
        const page2pagination = {
          count: 3,
          previous: '/elements/element1/children/?page=1',
          next: '/elements/element1/children/?page=3',
          number: 2
        }
        mock.onGet('/elements/element1/children/', { params: { page: 2 } }).reply(200, {
          ...page2pagination,
          results: [element3]
        })
        await store.dispatch('elements/nextChildren', { id: 'element1', max: 2 })

        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/nextChildren',
            payload: { id: 'element1', max: 2 }
          },
          {
            mutation: 'elements/addChildren',
            payload: { element: 'element1', children: [element2] }
          },
          {
            mutation: 'elements/setChildrenPagination',
            payload: { element: 'element1', ...page1pagination }
          },
          {
            mutation: 'elements/addChildren',
            payload: { element: 'element1', children: [element3] }
          },
          {
            mutation: 'elements/setChildrenPagination',
            payload: { element: 'element1', ...page2pagination }
          }
        ])
        assert.deepStrictEqual(store.state.elements.elements, { element1, element2, element3 })
        assert.deepStrictEqual(store.state.elements.links, {
          element1: {
            parents: [],
            children: [
              'element2',
              'element3'
            ]
          },
          element2: {
            parents: ['element1'],
            children: []
          },
          element3: {
            parents: ['element1'],
            children: []
          }
        })

        assert.deepStrictEqual(Object.keys(store.state.elements.childrenPagination), ['element1'])
        const actualPagination = store.state.elements.childrenPagination.element1
        // Ensure retrievedOn is a valid date string
        assert.ok(new Date(actualPagination.retrievedOn) instanceof Date)
        delete actualPagination.retrievedOn
        assert.deepStrictEqual(actualPagination, page2pagination)
      })

      describe('uses If-Modified-Since', () => {
        it('aborts on HTTP 304', async () => {
          mock.onGet('/elements/element1/children/').reply(304)

          const state = {
            elements: {
              element1: { id: 'element1', name: 'Element 1' }
            },
            links: {},
            childrenPagination: {
              element1: {
                count: 0,
                previous: null,
                next: null,
                number: 1,
                retrievedOn: 'Wed, 21 Oct 2015 07:28:00 GMT'
              }
            }
          }
          store.state.elements = state

          await store.dispatch('elements/nextChildren', { id: 'element1', max: Infinity })
          assert.strictEqual(mock.history.get.length, 1)
          assert.deepStrictEqual({ ...mock.history.get[0].headers }, {
            Accept: 'application/json, text/plain, */*',
            'Cache-Control': 'no-cache',
            'Content-Type': undefined,
            'If-Modified-Since': 'Wed, 21 Oct 2015 07:28:00 GMT'
          })
          assert.deepStrictEqual(store.history, [
            {
              action: 'elements/nextChildren',
              payload: { id: 'element1', max: Infinity }
            }
          ])
          assert.deepStrictEqual(store.state.elements, state)
        })

        it('restarts from scratch on HTTP 200', async () => {
          const element1 = {
            id: 'element1',
            name: 'Element 1'
          }
          const element2 = {
            id: 'element2',
            name: 'Element 2'
          }

          store.state.elements = {
            elements: {
              element1
            },
            links: {},
            childrenPagination: {
              element1: {
                count: 0,
                previous: null,
                next: null,
                number: 1,
                retrievedOn: 'Wed, 21 Oct 2015 07:28:00 GMT'
              }
            }
          }

          mock.onGet('/elements/element1/children/').reply(200, {
            count: 1,
            number: 1,
            results: [element2]
          })

          await store.dispatch('elements/nextChildren', { id: 'element1', max: Infinity })

          assert.deepStrictEqual(store.history, [
            {
              action: 'elements/nextChildren',
              payload: { id: 'element1', max: Infinity }
            },
            {
              mutation: 'elements/resetChildren',
              payload: { element: 'element1' }
            },
            {
              mutation: 'elements/addChildren',
              payload: { element: 'element1', children: [element2] }
            },
            {
              mutation: 'elements/setChildrenPagination',
              payload: { element: 'element1', count: 1, number: 1 }
            }
          ])

          assert.deepStrictEqual(store.state.elements.elements, {
            element1,
            element2
          })
          assert.deepStrictEqual(store.state.elements.links, {
            element1: {
              parents: [],
              children: ['element2']
            },
            element2: {
              parents: ['element1'],
              children: []
            }
          })

          assert.deepStrictEqual(Object.keys(store.state.elements.childrenPagination), ['element1'])
          const actualPagination = store.state.elements.childrenPagination.element1
          // Ensure retrievedOn is a valid date string
          assert.ok(new Date(actualPagination.retrievedOn) instanceof Date)
          delete actualPagination.retrievedOn
          assert.deepStrictEqual(actualPagination, { count: 1, number: 1 })
        })
      })

      it('handles with_has_children parameter', async () => {
        mock.onGet('/elements/element1/children/').reply(200, { count: 0, number: 1, results: [] })
        await store.dispatch('elements/nextChildren', { id: 'element1', max: 42, hasChildren: true })
        assert.strictEqual(mock.history.get.length, 1)
        assert.deepStrictEqual(mock.history.get[0].params, {
          page: 1,
          with_has_children: true
        })
      })

      it('handles errors', async () => {
        const element1 = { id: 'element1', name: 'Element 1' }
        const element2 = { id: 'element2', name: 'Element 2' }
        store.state.elements = {
          elements: {
            element1
          },
          links: {},
          childrenPagination: {}
        }
        const page1pagination = {
          count: 3,
          previous: null,
          next: '/elements/element1/children/?page=2',
          number: 1
        }
        mock.onGet('/elements/element1/children/', { params: { page: 1 } }).reply(200, {
          ...page1pagination,
          results: [element2]
        })
        mock.onGet('/elements/element1/children/', { params: { page: 2 } }).reply(500)

        await assertRejects(async () => {
          await store.dispatch('elements/nextChildren', { id: 'element1', max: Infinity })
        })

        assert.deepStrictEqual(store.history, [
          { action: 'elements/nextChildren', payload: { id: 'element1', max: Infinity } },
          { mutation: 'elements/addChildren', payload: { element: 'element1', children: [element2] } },
          { mutation: 'elements/setChildrenPagination', payload: { element: 'element1', ...page1pagination } },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
        assert.deepStrictEqual(store.state.elements.elements, { element1, element2 })
        assert.deepStrictEqual(store.state.elements.links, {
          element1: {
            parents: [],
            children: [
              'element2'
            ]
          },
          element2: {
            parents: ['element1'],
            children: []
          }
        })

        assert.deepStrictEqual(Object.keys(store.state.elements.childrenPagination), ['element1'])
        const actualPagination = store.state.elements.childrenPagination.element1
        // Ensure retrievedOn is a valid date string
        assert.ok(new Date(actualPagination.retrievedOn) instanceof Date)
        delete actualPagination.retrievedOn
        assert.deepStrictEqual(actualPagination, page1pagination)
      })
    })

    describe('getParentsBulk', () => {
      it('loads multiple pages', async () => {
        const element1 = { id: 'element1', name: 'Element 1' }
        const element2 = { id: 'element2', name: 'Element 2' }
        const element3 = { id: 'element3', name: 'Element 3' }
        store.state.elements = {
          elements: {
            element1
          },
          parents: {}
        }
        const page1pagination = {
          count: 3,
          previous: null,
          next: '/elements/element1/parents/?page=2',
          number: 1
        }
        mock.onGet('/elements/element1/parents/', { params: { page: 1, with_corpus: false, with_zone: false, recursive: true } }).reply(200, {
          ...page1pagination,
          results: [element2]
        })
        const page2pagination = {
          count: 3,
          previous: '/elements/element1/parents/?page=1',
          next: null,
          number: 2
        }
        mock.onGet('/elements/element1/parents/', { params: { page: 2, with_corpus: false, with_zone: false, recursive: true } }).reply(200, {
          ...page2pagination,
          results: [element3]
        })
        await store.dispatch('elements/getParentsBulk', { id: 'element1', corpus: { id: 'corpusid' } })

        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/getParentsBulk',
            payload: { id: 'element1', corpus: { id: 'corpusid' } }
          },
          {
            mutation: 'elements/addParents',
            payload: { element: 'element1', parents: [element2], corpus: { id: 'corpusid' } }
          },
          {
            mutation: 'elements/addParents',
            payload: { element: 'element1', parents: [element3], corpus: { id: 'corpusid' } }
          }
        ])
        assert.deepStrictEqual(store.state.elements.elements, {
          element1,
          element2: {
            ...element2,
            corpus: { id: 'corpusid' }
          },
          element3: {
            ...element3,
            corpus: { id: 'corpusid' }
          }
        })
        assert.deepStrictEqual(store.state.elements.parents, {
          element1: ['element2', 'element3']
        })
      })

      it('handles parameters', async () => {
        mock.onGet('/elements/element1/parents/').reply(200, { count: 0, number: 1, results: [] })
        await store.dispatch('elements/getParentsBulk', { id: 'element1', corpus: { id: 'corpusid' } })
        assert.strictEqual(mock.history.get.length, 1)
        assert.deepStrictEqual(mock.history.get[0].params, {
          page: 1,
          with_corpus: false,
          with_zone: false,
          recursive: true
        })
      })

      it('requires a corpus', async () => {
        const err = await assertRejects(async () => store.dispatch('elements/getParentsBulk', { id: 'elementid' }))
        assert.strictEqual(err.toString(), 'Error: A corpus is required.')
      })

      it('handles errors', async () => {
        const element1 = { id: 'element1', name: 'Element 1' }
        const element2 = { id: 'element2', name: 'Element 2' }
        store.state.elements = {
          elements: {
            element1
          },
          parents: {}
        }
        const page1pagination = {
          count: 3,
          previous: null,
          next: '/elements/element1/parents/?page=2',
          number: 1
        }
        mock.onGet('/elements/element1/parents/', { params: { page: 1, with_corpus: false, with_zone: false, recursive: true } }).reply(200, {
          ...page1pagination,
          results: [element2]
        })
        mock.onGet('/elements/element1/parents/', { params: { page: 2, with_corpus: false, with_zone: false, recursive: true } }).reply(500)

        const err = await assertRejects(async () => {
          await store.dispatch('elements/getParentsBulk', { id: 'element1', corpus: { id: 'corpusid' } })
        })
        assert.strictEqual(err.toString(), 'Error: Request failed with status code 500')

        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/getParentsBulk',
            payload: { id: 'element1', corpus: { id: 'corpusid' } }
          },
          {
            mutation: 'elements/addParents',
            payload: { element: 'element1', parents: [element2], corpus: { id: 'corpusid' } }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])
        assert.deepStrictEqual(store.state.elements.elements, {
          element1,
          element2: {
            ...element2,
            corpus: { id: 'corpusid' }
          }
        })
        assert.deepStrictEqual(store.state.elements.parents, {
          element1: ['element2']
        })
      })
    })

    describe('listNeighbors', () => {
      it("lists an element's neighbors", async () => {
        mock.onGet('/elements/elementid/neighbors/').reply(200, elementNeighborsSample)

        await store.dispatch('elements/listNeighbors', { id: 'elementid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/listNeighbors',
            payload: { id: 'elementid' }
          },
          {
            mutation: 'elements/addNeighbors',
            payload: { element: 'elementid', neighbors: elementNeighborsSample }
          }
        ])

        assert.deepStrictEqual(store.state.elements.neighbors, { elementid: elementNeighborsSample })
      })

      it('handles errors', async () => {
        mock.onGet('/elements/elementid/neighbors/').reply(500)

        await store.dispatch('elements/listNeighbors', { id: 'elementid' })

        assert.deepStrictEqual(store.history, [
          {
            action: 'elements/listNeighbors',
            payload: { id: 'elementid' }
          },
          {
            mutation: 'elements/addNeighbors',
            payload: { element: 'elementid', neighbors: null }
          },
          {
            mutation: 'notifications/notify',
            payload: {
              type: 'error',
              text: 'Request failed with status code 500'
            }
          }
        ])

        assert.deepStrictEqual(store.state.elements.neighbors, { elementid: null })
      })
    })
  })

  describe('getters', () => {
    describe('canWrite', () => {
      it('returns true if the user has write access on an element', () => {
        const getter = getters.canWrite({
          elements: {
            elementid: {
              rights: ['read', 'write']
            }
          }
        })
        assert.ok(getter('elementid'))
      })

      it('returns false if the user does not have write access', () => {
        const getter = getters.canWrite({
          elements: {
            elementid: {
              rights: ['read', 'admin']
            }
          }
        })
        assert.ok(!getter('elementid'))
      })

      it('returns false if there are no access rights on the element', () => {
        const getter = getters.canWrite({
          elements: {
            elementid: {
              id: 'elementid'
            }
          }
        })
        assert.ok(!getter('elementid'))
      })

      it('returns false for nonexistent elements', () => {
        const getter = getters.canWrite({ elements: {} })
        assert.ok(!getter('elementid'))
      })
    })

    describe('canAdmin', () => {
      it('returns true if the user has admin access on an element', () => {
        const getter = getters.canAdmin({
          elements: {
            elementid: {
              rights: ['read', 'admin']
            }
          }
        })
        assert.ok(getter('elementid'))
      })

      it('returns false if the user does not have admin access', () => {
        const getter = getters.canAdmin({
          elements: {
            elementid: {
              rights: ['read', 'write']
            }
          }
        })
        assert.ok(!getter('elementid'))
      })

      it('returns false if there are no access rights on the element', () => {
        const getter = getters.canAdmin({
          elements: {
            elementid: {
              id: 'elementid'
            }
          }
        })
        assert.ok(!getter('elementid'))
      })

      it('returns false for nonexistent elements', () => {
        const getter = getters.canAdmin({ elements: {} })
        assert.ok(!getter('elementid'))
      })
    })
  })
})
