import { assert } from 'chai'
import { createPinia, setActivePinia } from 'pinia'
import { useDisplayStore } from '@/stores'

describe('display', () => {
  let store

  before(() => {
    setActivePinia(createPinia())
    store = useDisplayStore()
  })

  beforeEach(() => {
    store.$reset()
  })

  describe('toggleElementDetails', () => {
    it('toggles without a value', () => {
      assert.ok(store.displayDetails)
      store.toggleElementDetails()
      assert.ok(!store.displayDetails)
    })

    it('sets directly with a boolean value', () => {
      assert.ok(store.displayDetails)
      store.toggleElementDetails(true)
      assert.ok(store.displayDetails)
      store.toggleElementDetails(false)
      assert.ok(!store.displayDetails)
    })
  })

  describe('toggleFolderTree', () => {
    it('toggles without a value', () => {
      assert.ok(!store.displayFolderTree)
      store.toggleFolderTree()
      assert.ok(store.displayFolderTree)
    })

    it('sets directly with a boolean value', () => {
      assert.ok(!store.displayFolderTree)
      store.toggleFolderTree(false)
      assert.ok(!store.displayFolderTree)
      store.toggleFolderTree(true)
      assert.ok(store.displayFolderTree)
    })
  })

  describe('toggleAnnotationsTree', () => {
    it('toggles without a value', () => {
      assert.ok(store.displayAnnotationsTree)
      store.toggleAnnotationsTree()
      assert.ok(!store.displayAnnotationsTree)
    })

    it('sets directly with a boolean value', () => {
      assert.ok(store.displayAnnotationsTree)
      store.toggleAnnotationsTree(false)
      assert.ok(!store.displayAnnotationsTree)
      store.toggleAnnotationsTree(true)
      assert.ok(store.displayAnnotationsTree)
    })
  })

  describe('toggleEntityTypes', () => {
    it('toggles without a value', () => {
      assert.ok(store.displayEntityTypes)
      store.toggleEntityTypes()
      assert.ok(!store.displayEntityTypes)
    })

    it('sets directly with a boolean value', () => {
      assert.ok(store.displayEntityTypes)
      store.toggleEntityTypes(false)
      assert.ok(!store.displayEntityTypes)
      store.toggleEntityTypes(true)
      assert.ok(store.displayEntityTypes)
    })
  })

  describe('toggleElementClasses', () => {
    it('toggles without a value', () => {
      assert.ok(!store.displayElementClasses)
      store.toggleElementClasses()
      assert.ok(store.displayElementClasses)
    })

    it('sets directly with a boolean value', () => {
      assert.ok(!store.displayElementClasses)
      store.toggleElementClasses(true)
      assert.ok(store.displayElementClasses)
      store.toggleElementClasses(false)
      assert.ok(!store.displayElementClasses)
    })
  })

  describe('toggleCompactDisplay', () => {
    it('toggles without a value', () => {
      assert.ok(!store.compactDisplay)
      store.toggleCompactDisplay()
      assert.ok(store.compactDisplay)
    })

    it('sets directly with a boolean value', () => {
      assert.ok(!store.compactDisplay)
      store.toggleCompactDisplay(true)
      assert.ok(store.compactDisplay)
      store.toggleCompactDisplay(false)
      assert.ok(!store.compactDisplay)
    })
  })

  describe('toggleAllPaths', () => {
    it('toggles without a value', () => {
      assert.ok(!store.displayAllPaths)
      store.toggleAllPaths()
      assert.ok(store.displayAllPaths)
    })

    it('sets directly with a boolean value', () => {
      store.displayAllPaths = false
      assert.ok(!store.displayAllPaths)
      store.toggleAllPaths(true)
      assert.ok(store.displayAllPaths)
      store.toggleAllPaths(false)
      assert.ok(!store.displayAllPaths)
    })
  })

  describe('toggleDropdown', () => {
    it('toggles without a value', () => {
      assert.deepStrictEqual(store.dropdowns, {})
      store.toggleDropdown('something')
      assert.deepStrictEqual(store.dropdowns, { something: true })
      store.toggleDropdown('something')
      assert.deepStrictEqual(store.dropdowns, { something: false })
    })

    it('sets directly with a boolean value', () => {
      assert.deepStrictEqual(store.dropdowns, {})
      store.toggleDropdown('something', false)
      assert.deepStrictEqual(store.dropdowns, { something: false })
    })
  })

  describe('setPageSize', () => {
    it('updates the navigation page size', () => {
      assert.strictEqual(store.navigationPageSize, null)
      store.setPageSize(50)
      assert.strictEqual(store.navigationPageSize, 50)
    })

    it('only accepts positive integers', () => {
      const invalid = [
        null,
        undefined,
        false,
        NaN,
        Infinity,
        -Infinity,
        0,
        -4,
        1.2,
        ''
      ]
      for (const value of invalid) {
        assert.strictEqual(store.navigationPageSize, null)
        assert.throws(
          () => store.setPageSize(value),
          TypeError,
          'Page size must be a positive integer'
        )
        assert.strictEqual(store.navigationPageSize, null)
      }
    })
  })

  describe('setLastMetadata', () => {
    it('sets the last metadata name and type', () => {
      assert.strictEqual(store.lastMetadataName, null)
      assert.strictEqual(store.lastMetadataType, null)
      store.setLastMetadata('blah', 'numeric')
      assert.strictEqual(store.lastMetadataName, 'blah')
      assert.strictEqual(store.lastMetadataType, 'numeric')
    })
  })

  it('detects the screen size on reset', () => {
    // The screen size is supposed to be set to 1024×768 in unit tests
    assert.strictEqual(window.innerWidth, 1024)
    assert.strictEqual(window.innerHeight, 768)

    store.screen = {
      width: 42,
      height: 42
    }
    store.$reset()

    assert.deepStrictEqual(store.screen, {
      width: 1024,
      height: 768
    })
  })
})
