import { assert } from 'chai'
import axios from 'axios'
import { FakeAxios, setUpTestPinia, actionsCompletedPlugin } from '../testhelpers'
import { jobsSample, exportSample, exportSourcesSample } from '../samples.js'
import { useExportStore, useJobsStore, useNotificationStore } from '@/stores'

describe('exports', () => {
  describe('actions', () => {
    let pinia, mock, store, notificationStore, jobsStore

    before('Setting up mocks', () => {
      [pinia] = setUpTestPinia([actionsCompletedPlugin])
      mock = new FakeAxios(axios)
      store = useExportStore()
      notificationStore = useNotificationStore()
      jobsStore = useJobsStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('list', () => {
      it('lists exports in a corpus', async () => {
        mock.onGet('/corpus/corpusid/export/').reply(200, { count: 1, results: [exportSample] })
        await store.list('corpusid')
        assert.deepStrictEqual(store.corpusExports, {
          corpusid: {
            count: 1,
            results: [exportSample]
          }
        })
      })

      it('handles errors', async () => {
        mock.onGet('/corpus/corpusid/export/').reply(400, { detail: 'oh no' })
        await store.list('corpusid')
        assert.deepStrictEqual(store.corpusExports, {})
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'oh no'
          }
        ])
      })
    })

    describe('start', () => {
      it('starts an export', async () => {
        mock.onPost('/corpus/corpusid/export/').reply(201, exportSample)
        mock.onGet('/jobs/').reply(200, jobsSample)

        await store.start('corpusid')
        await pinia.actionsCompleted()

        assert.deepStrictEqual(jobsStore.jobs, {
          jobid: jobsSample[0]
        })
      })

      it('handles errors', async () => {
        mock.onPost('/corpus/corpusid/export/').reply(402, { detail: 'not enough bitcoins' })

        await store.start('corpusid')
        await store.actionsCompleted()

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'not enough bitcoins'
          }
        ])
      })
    })

    describe('delete', () => {
      it('deletes an export', async () => {
        store.corpusExports = {
          corpusid: { count: 1, results: [exportSample] }
        }

        mock.onDelete('/export/exportid/').reply(204)
        mock.onGet('/corpus/corpusid/export/').reply(200, {})

        await store.delete('exportid', 'corpusid')
        await store.actionsCompleted()

        assert.deepStrictEqual(store.corpusExports.corpusid, {})
      })

      it('handles errors', async () => {
        store.corpusExports = {
          corpusid: { count: 1, results: [exportSample] }
        }

        mock.onDelete('/export/exportid/').reply(400, { detail: 'no' })

        await store.delete('exportid', 'corpusid')
        await store.actionsCompleted()

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'no'
          }
        ])
      })
    })

    describe('listSources', () => {
      it('lists export sources', async () => {
        mock.onGet('/export-sources/').reply(200, exportSourcesSample)

        await store.listSources()

        assert.deepStrictEqual(store.sources, exportSourcesSample)
      })

      it('handles errors', async () => {
        mock.onGet('/export-sources/').reply(400, { detail: 'oh no' })

        await store.listSources()

        assert.strictEqual(store.sources, null)
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'oh no'
          }
        ])
      })
    })
  })
})
